#include <string.h>
#include <stdio.h>
#include "COP_FPU.h"
#include "MIPS.h"

CCOP_FPU			g_COPFPU(MIPS_REGSIZE_64);

uint8				CCOP_FPU::m_nFT;
uint8				CCOP_FPU::m_nFS;
uint8				CCOP_FPU::m_nFD;

uint32				CCOP_FPU::m_nCCMask[8] =
{
	0x00800000,
	0x02000000,
	0x04000000,
	0x08000000,
	0x10000000,
	0x20000000,
	0x40000000,
	0x80000000
};

CCOP_FPU::CCOP_FPU(MIPS_REGSIZE nRegSize) :
CMIPSCoprocessor(nRegSize)
{

}

void CCOP_FPU::CompileInstruction(uint32 nAddress, CCacheBlock* pBlock, CMIPS* pCtx, bool nParent)
{
	if(nParent)
	{
		SetupQuickVariables(nAddress, pBlock, pCtx);
	}

	m_nFT			= (uint8)((m_nOpcode >> 16) & 0x1F);
	m_nFS			= (uint8)((m_nOpcode >> 11) & 0x1F);
	m_nFD			= (uint8)((m_nOpcode >> 6) & 0x1F);

	switch((m_nOpcode >> 26) & 0x3F)
	{
	case 0x11:
		//COP1
		m_pOpGeneral[(m_nOpcode >> 21) & 0x1F]();
		break;
	case 0x31:
		LWC1();
		break;
	case 0x39:
		SWC1();
		break;
	default:
		Illegal();
		break;
	}
}

void CCOP_FPU::SetCCBit(bool nCondition, uint32 nMask)
{
	m_pB->PushTop();
	
	m_pB->BeginJcc(nCondition);
	{
		m_pB->PushAddr(&m_pCtx->m_State.nFCSR);
		m_pB->OrImm(nMask);
		m_pB->PullAddr(&m_pCtx->m_State.nFCSR);
	}
	m_pB->EndJcc();

	m_pB->BeginJcc(!nCondition);
	{
		m_pB->PushAddr(&m_pCtx->m_State.nFCSR);
		m_pB->AndImm(~nMask);
		m_pB->PullAddr(&m_pCtx->m_State.nFCSR);
	}
	m_pB->EndJcc();
}

void CCOP_FPU::TestCCBit(uint32 nMask)
{
	m_pB->PushAddr(&m_pCtx->m_State.nFCSR);
	m_pB->AndImm(nMask);
	m_pB->PushImm(0);
	m_pB->Cmp(JCC_CONDITION_NE);
}

//////////////////////////////////////////////////
//General Opcodes
//////////////////////////////////////////////////

//00
void CCOP_FPU::MFC1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	SignExtendTop32(m_nFT);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nFT]);
}

//04
void CCOP_FPU::MTC1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nFT]);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
}

//08
void CCOP_FPU::BC1()
{
	switch((m_nOpcode >> 16) & 0x03)
	{
	case 0x00:
		BC1F();
		break;
	case 0x01:
		BC1T();
		break;
	case 0x02:
		BC1FL();
		break;
	case 0x03:
		BC1TL();
		break;
	default:
		Illegal();
		break;
	}
}

//10
void CCOP_FPU::S()
{
	m_pOpSingle[(m_nOpcode & 0x3F)]();
}

//14
void CCOP_FPU::W()
{
	m_pOpWord[(m_nOpcode & 0x3F)]();
}

//////////////////////////////////////////////////
//Branch Opcodes
//////////////////////////////////////////////////

//00
void CCOP_FPU::BC1F()
{
	TestCCBit(m_nCCMask[(m_nOpcode >> 18) & 0x07]);
	Branch(false);
}

//01
void CCOP_FPU::BC1T()
{
	TestCCBit(m_nCCMask[(m_nOpcode >> 18) & 0x07]);
	Branch(true);
}

//02
void CCOP_FPU::BC1FL()
{
	TestCCBit(m_nCCMask[(m_nOpcode >> 18) & 0x07]);
	BranchLikely(false);
}

//03
void CCOP_FPU::BC1TL()
{
	TestCCBit(m_nCCMask[(m_nOpcode >> 18) & 0x07]);
	BranchLikely(true);
}

//////////////////////////////////////////////////
//Single Precision Floating Point Opcodes
//////////////////////////////////////////////////

//00
void CCOP_FPU::ADD_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_Add();
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//01
void CCOP_FPU::SUB_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_Sub();
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//02
void CCOP_FPU::MUL_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_Mul();
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//03
void CCOP_FPU::DIV_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_Div();
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//06
void CCOP_FPU::MOV_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//07
void CCOP_FPU::NEG_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_Neg();
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//24
void CCOP_FPU::CVT_W_S()
{
	//Load the rounding mode from FCSR?
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PullAddrWord(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//32
void CCOP_FPU::C_EQ_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_Cmp(JCC_CONDITION_EQ);

	SetCCBit(true, m_nCCMask[((m_nOpcode >> 8) & 0x07)]);
}

//34
void CCOP_FPU::C_LT_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_Cmp(JCC_CONDITION_BL);

	SetCCBit(true, m_nCCMask[((m_nOpcode >> 8) & 0x07)]);
}

//36
void CCOP_FPU::C_LE_S()
{
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->FPU_PushAddrSingle(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_Cmp(JCC_CONDITION_BE);

	SetCCBit(true, m_nCCMask[((m_nOpcode >> 8) & 0x07)]);
}

//////////////////////////////////////////////////
//Word Sized Integer Opcodes
//////////////////////////////////////////////////

//20
void CCOP_FPU::CVT_S_W()
{
	m_pB->FPU_PushAddrWord(&m_pCtx->m_State.nCOP10[m_nFS * 2]);
	m_pB->FPU_PullAddrSingle(&m_pCtx->m_State.nCOP10[m_nFD * 2]);
}

//////////////////////////////////////////////////
//Miscellanous Opcodes
//////////////////////////////////////////////////

//31
void CCOP_FPU::LWC1()
{
	ComputeMemAccessAddr();

	//Load the word
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 2, true);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
}

//39
void CCOP_FPU::SWC1()
{
	ComputeMemAccessAddr();

	//Write the words
	m_pB->PushAddr(&m_pCtx->m_State.nCOP10[m_nFT * 2]);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 3, false);
}

//////////////////////////////////////////////////
//Opcode Tables
//////////////////////////////////////////////////

void (*CCOP_FPU::m_pOpGeneral[0x20])() = 
{
	//0x00
	MFC1,			Illegal,		Illegal,		Illegal,		MTC1,			Illegal,		Illegal,		Illegal,
	//0x08
	BC1,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	S,				Illegal,		Illegal,		Illegal,		W,				Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CCOP_FPU::m_pOpSingle[0x40])() =
{
	//0x00
	ADD_S,			SUB_S,			MUL_S,			DIV_S,			Illegal,		Illegal,		MOV_S,			NEG_S,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x20
	Illegal,		Illegal,		Illegal,		Illegal,		CVT_W_S,		Illegal,		Illegal,		Illegal,
	//0x28
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x30
	Illegal,		Illegal,		C_EQ_S,			Illegal,		C_LT_S,			Illegal,		C_LE_S,			Illegal,
	//0x38
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CCOP_FPU::m_pOpWord[0x40])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x20
	CVT_S_W,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x28
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x30
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x38
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void CCOP_FPU::GetInstruction(uint32 nOpcode, char* sText)
{
	switch((nOpcode >> 26) & 0x3F)
	{
	case 0x11:
		//COP1
		switch((nOpcode >> 21) & 0x1F)
		{
		case 0x00:
			strcpy(sText, "MFC1");
			break;
		case 0x04:
			strcpy(sText, "MTC1");
			break;
		case 0x08:
			switch((nOpcode >> 16) & 0x03)
			{
			case 0x00:
				strcpy(sText, "BC1F");
				break;
			case 0x01:
				strcpy(sText, "BC1T");
				break;
			case 0x02:
				strcpy(sText, "BC1FL");
				break;
			case 0x03:
				strcpy(sText, "BC1TL");
				break;
			default:
				strcpy(sText, "???");
				break;
			}
			break;
		case 0x10:
			switch(nOpcode & 0x3F)
			{
			case 0x00:
				strcpy(sText, "ADD.S");
				break;
			case 0x01:
				strcpy(sText, "SUB.S");
				break;
			case 0x02:
				strcpy(sText, "MUL.S");
				break;
			case 0x03:
				strcpy(sText, "DIV.S");
				break;
			case 0x06:
				strcpy(sText, "MOV.S");
				break;
			case 0x07:
				strcpy(sText, "NEG.S");
				break;
			case 0x24:
				strcpy(sText, "CVT.W.S");
				break;
			case 0x32:
				strcpy(sText, "C.EQ.S");
				break;
			case 0x34:
				strcpy(sText, "C.LT.S");
				break;
			case 0x36:
				strcpy(sText, "C.LE.S");
				break;
			default:
				strcpy(sText, "???");
				break;
			}
			break;
		case 0x14:
			switch(nOpcode & 0x3F)
			{
			case 0x20:
				strcpy(sText, "CVT.S.W");
				break;
			default:
				strcpy(sText, "???");
				break;
			}
			break;
		default:
			strcpy(sText, "???");
			break;
		}
		break;
	case 0x31:
		strcpy(sText, "LWC1");
		break;
	case 0x39:
		strcpy(sText, "SWC1");
		break;
	default:
		strcpy(sText, "???");
		break;
	}
}

void CCOP_FPU::GetArguments(uint32 nAddress, uint32 nOpcode, char* sText)
{
	//Notes for FPU
	//RT = FT
	//RD = FS
	//SA = FD
	uint8 nFT, nFD, nFS, nCC;
	uint16 nImm;
	nFT		= (uint8) ((nOpcode >> 16) & 0x1F);
	nFS		= (uint8) ((nOpcode >> 11) & 0x1F);
	nFD		= (uint8) ((nOpcode >> 6) & 0x1F);
	nCC		= (uint8) ((nOpcode >> 8) & 0x07);
	nImm	= (uint16)(nOpcode & 0xFFFF);

	switch((nOpcode >> 26) & 0x3F)
	{
	case 0x11:
		//COP1
		switch((nOpcode >> 21) & 0x1F)
		{
		case 0x00:
		case 0x04:
			sprintf(sText, "%s, F%i", CMIPS::m_sGPRName[nFT], nFS);
			break;
		case 0x08:
			sprintf(sText, "CC%i, $%0.8X", (nOpcode >> 18) & 0x07, nAddress + CMIPS::GetBranch(nImm) + 4);
			break;
		case 0x10:
			switch(nOpcode & 0x3F)
			{
			case 0x00:
			case 0x01:
			case 0x02:
			case 0x03:
				sprintf(sText, "F%i, F%i, F%i", nFD, nFS, nFT);
				break;
			case 0x06:
			case 0x07:
			case 0x24:
				sprintf(sText, "F%i, F%i", nFD, nFS);
				break;
			case 0x32:
			case 0x34:
			case 0x36:
				sprintf(sText, "CC%i, F%i, F%i", nCC, nFS, nFT);
				break;
			default:
				strcpy(sText, "");
				break;
			}
			break;
		case 0x14:
			switch(nOpcode & 0x3F)
			{
			case 0x20:
				sprintf(sText, "F%i, F%i", nFD, nFS);
				break;
			default:
				strcpy(sText, "");
				break;
			}
			break;
		default:
			strcpy(sText, "");
			break;
		}
		break;
	case 0x31:
	case 0x39:
		sprintf(sText, "F%i, $%0.4X(%s)", nFT, nImm, CMIPS::m_sGPRName[(nOpcode >> 21) & 0x1F]);
		break;
	default:
		strcpy(sText, "");
		break;
	}

}

uint32 CCOP_FPU::GetEffectiveAddress(uint32 nAddress, uint32 nOpcode)
{
	return 0;
}

bool CCOP_FPU::IsBranch(uint32 nOpcode)
{
	switch((nOpcode >> 26) & 0x3F)
	{
	case 0x11:
		//COP1
		switch((nOpcode >> 21) & 0x1F)
		{
		case 0x08:
			return true;
			break;
		}
		break;
	}
	return false;
}
