#include <stdio.h>
#include "PS2VM.h"
#include "DMAC.h"
#include "INTC.h"
#include "GIF.h"
#include "SIF.h"

using namespace Framework;

uint32		CDMAC::m_D_STAT = 0;

uint32		CDMAC::m_D2_CHCR = 0;
uint32		CDMAC::m_D2_MADR = 0;
uint32		CDMAC::m_D2_SIZE = 0;

uint32		CDMAC::m_D5_CHCR = 0;
uint32		CDMAC::m_D5_MADR = 0;
uint32		CDMAC::m_D5_QWC = 0;

uint32		CDMAC::m_D6_CHCR	= 0;
uint32		CDMAC::m_D6_MADR	= 0;
uint32		CDMAC::m_D6_QWC		= 0;
uint32		CDMAC::m_D6_TADR	= 0;


//DMA channels (EE side)
//0 - VIF0
//1 - VIF1
//2 - GIF
//3 - IPU (incoming)
//4 - IPU (outgoing)
//5 - SIF0 (from IOP)
//6 - SIF1 (to IOP?)
//7 - SIF2
//8 - SPR (incoming)
//9 - SPR (outgoing)

void CDMAC::Reset()
{
	m_D_STAT = 0;
}

bool CDMAC::IsInterruptPending()
{
	uint16 nMask, nStatus;
	nMask	= (uint16)((m_D_STAT & 0x63FF0000) >> 16);
	nStatus = (uint16)((m_D_STAT & 0x0000E3FF) >>  0);

	return ((nMask & nStatus) != 0);
}

uint32 CDMAC::GetRegister(uint32 nAddress)
{
	switch(nAddress)
	{
	case 0x1000A000:
		return m_D2_CHCR;
		break;

	case 0x1000A004:
	case 0x1000A008:
	case 0x1000A00C:
		return 0;
		break;

	case 0x1000E010:
		return m_D_STAT;
		break;

	default:
		printf("DMAC: Read to an unhandled IO port (0x%0.8X).\r\n", nAddress);
		break;
	}

	DisassembleGet(nAddress);

	return 0;
}

void CDMAC::SetRegister(uint32 nAddress, uint32 nData)
{
	switch(nAddress)
	{
	//D2_CHCR
	case 0x1000A000:
		m_D2_CHCR = nData;
		if(m_D2_CHCR & 0x100)
		{
			CGIF::ExecuteDMA(m_D2_MADR, m_D2_SIZE * 0x10);
			m_D2_CHCR &= ~0x100;
		}
		break;
	case 0x1000A004:
	case 0x1000A008:
	case 0x1000A00C:
		break;

	//D2_MADR
	case 0x1000A010:
		m_D2_MADR = nData;
		break;
	case 0x1000A014:
	case 0x1000A018:
	case 0x1000A01C:
		break;

	//D2_SIZE
	case 0x1000A020:
		m_D2_SIZE = nData;
		break;
	case 0x1000A024:
	case 0x1000A028:
	case 0x1000A02C:
		break;

	//D5_CHCR
	case 0x1000C000:
		m_D5_CHCR = nData;
		if(m_D5_CHCR & 0x100)
		{
			memcpy(CPS2VM::m_pRAM + m_D5_MADR, CSIF::m_pRAM, m_D5_QWC * 0x10);
			m_D5_CHCR	&= ~0x100;
			m_D_STAT	|= 0x20;
			if(IsInterruptPending())
			{
				CINTC::CheckInterrupts();
			}
		}
		break;
	case 0x1000C004:
	case 0x1000C008:
	case 0x1000C00C:
		break;

	//D5_MADR
	case 0x1000C010:
		m_D5_MADR = nData;
		break;
	case 0x1000C014:
	case 0x1000C018:
	case 0x1000C01C:
		break;

	//D5_QWC
	case 0x1000C020:
		m_D5_QWC = nData;
	case 0x1000C024:
	case 0x1000C028:
	case 0x1000C02C:
		break;

	//D6_CHCR
	case 0x1000C400:
		m_D6_CHCR = nData;
		if(m_D6_CHCR & 0x100)
		{
			CSIF::ReceiveDMA(m_D6_MADR, m_D6_TADR, m_D6_QWC * 0x10);
			m_D6_CHCR &= ~0x100;
		}
		break;
	case 0x1000C404:
	case 0x1000C408:
	case 0x1000C40C:
		break;

	//D6_MADR
	case 0x1000C410:
		m_D6_MADR = nData;
		break;
	case 0x1000C414:
	case 0x1000C418:
	case 0x1000C41C:
		break;

	//D6_QWC
	case 0x1000C420:
		m_D6_QWC = nData;
		break;
	case 0x1000C424:
	case 0x1000C428:
	case 0x1000C42C:
		break;

	//D6_TADR
	case 0x1000C430:
		m_D6_TADR = nData;
		break;
	case 0x1000C434:
	case 0x1000C438:
	case 0x1000C43C:
		break;

	case 0x1000E010:
		uint32 nStat, nMask;
		nStat = nData & 0x0000FFFF;
		nMask = nData & 0xFFFF0000;

		//Set the masks
//		m_D_STAT &= 0x0000FFFF;
//		m_D_STAT |= nMask;
		m_D_STAT ^= nMask;

		//Clear the interrupt status
		m_D_STAT &= ~nStat;

		//Trigger interrupts
		if(IsInterruptPending())
		{
			CINTC::CheckInterrupts();
		}

		break;
	case 0x1000E014:
	case 0x1000E018:
	case 0x1000E01C:
		break;

	default:
		printf("DMAC: Wrote to an unhandled IO port (0x%0.8X, 0x%0.8X).\r\n", nAddress, nData);
		break;
	}

	DisassembleSet(nAddress, nData);
}

void CDMAC::LoadState(CStream* pStream)
{
	pStream->Read(&m_D_STAT, 4);
}

void CDMAC::SaveState(CStream* pStream)
{
	pStream->Write(&m_D_STAT, 4);
}

void CDMAC::DisassembleGet(uint32 nAddress)
{
	if(!CPS2VM::m_Logging.GetDMACLoggingStatus()) return;

	switch(nAddress)
	{
	case 0x1000A000:
		printf("DMAC: = D2_CHCR.\r\n");
		break;
	case 0x1000E010:
		printf("DMAC: = D_STAT.\r\n");
		break;
	}
}

void CDMAC::DisassembleSet(uint32 nAddress, uint32 nData)
{
	if(!CPS2VM::m_Logging.GetDMACLoggingStatus()) return;

	switch(nAddress)
	{
	case 0x1000A000:
		printf("DMAC: D2_CHCR = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000A010:
		printf("DMAC: D2_MADR = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000A020:
		printf("DMAC: D2_SIZE = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C000:
		printf("DMAC: D5_CHCR = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C010:
		printf("DMAC: D5_MADR = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C020:
		printf("DMAC: D5_QWC = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C400:
		printf("DMAC: D6_CHCR = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C410:
		printf("DMAC: D6_MADR = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C420:
		printf("DMAC: D6_QWC = 0x%0.8X.\r\n", nData);
		break;
	case 0x1000C430:
		printf("DMAC: D6_TADR = 0x%0.8X.\r\n", nData);
		break;
	}
}
