#ifndef _MIPS_H_
#define _MIPS_H_

#include "Types.h"
#include "MemoryMap.h"
#include "ExecMap.h"
#include "MIPSArchitecture.h"
#include "MIPSCoprocessor.h"
#include "MIPSAnalysis.h"
#include "MIPSTags.h"

struct MIPSSTATE
{
	uint32		nPC;
	uint32		nDelayedJumpAddr;
	uint32		nGPR0[32];
	uint32		nGPR1[32];
	uint32		nGPR2[32];
	uint32		nGPR3[32];
	uint32		nHI[2];
	uint32		nLO[2];
	uint32		nHI1[2];
	uint32		nLO1[2];
	uint32		nCOP0[32];
	uint32		nCOP10[32];
	uint32		nCOP11[32];
	uint32		nFCSR;
};

#define MIPS_INVALID_PC		(0x00000001)

typedef unsigned int	(*TickFunction)(unsigned int);
typedef void			(*SysCallHandler)();
typedef uint32			(*AddressTranslator)(CMIPS*, uint32, uint32);

class CMIPS
{
public:
								CMIPS(MEMORYMAP_ENDIANESS, uint32, uint32);
								~CMIPS();
	void						ToggleBreakpoint(uint32);
	void						InvalidateCache();
	void						Step();
	RET_CODE					Execute(int);
	bool						MustBreak();
	bool						IsBranch(uint32);
	static long					GetBranch(uint16);
	static uint32				TranslateAddress64(CMIPS*, uint32, uint32);
	static void					DefaultSysCallHandler();
	
	MIPSSTATE					m_State;
	uint32						m_nIllOpcode;

	int							m_nQuota;
	CMIPSArchitecture*			m_pArch;
	CMIPSCoprocessor*			m_pCOP[4];
	CMemoryMap*					m_pMemoryMap;
	CExecMap*					m_pExecMap;
	Framework::CList<void>		m_Breakpoint;

	CMIPSAnalysis*				m_pAnalysis;
	CMIPSTags					m_Comments;
	CMIPSTags					m_Functions;

	AddressTranslator			m_pAddrTranslator;
	SysCallHandler				m_pSysCallHandler;
	TickFunction				m_pTickFunction;

	static const char*			m_sGPRName[];
};

#endif
