#include <stdio.h>
#include "PS2VM.h"
#include "DMAC.h"
#include "INTC.h"
#include "GIF.h"
#include "SIF.h"
#include "MA_EE.h"
#include "COP_SCU.h"
#include "COP_FPU.h"
#include "PtrMacro.h"
#include "StdStream.h"

#ifdef _DEBUG

#define		SCREENTICKS		10000
#define		VBLANKTICKS		1000

#else

//#define		SCREENTICKS		50000000
#define		SCREENTICKS		500000
#define		VBLANKTICKS		10000

#endif

using namespace Framework;

uint8*			CPS2VM::m_pRAM						= NULL;
uint8*			CPS2VM::m_pBIOS						= NULL;
pthread_t		CPS2VM::m_nThread					= 0;
CMIPS			CPS2VM::m_EE(MEMORYMAP_ENDIAN_LSBF, 0x00000000, 0x20000000);
CThreadMsg		CPS2VM::m_MsgBox;
PS2VM_STATUS	CPS2VM::m_nStatus					= PS2VM_STATUS_PAUSED;
CGSHandler*		CPS2VM::m_pGS						= NULL;
CPadHandler*	CPS2VM::m_pPad						= NULL;
unsigned int	CPS2VM::m_nVBlankTicks				= SCREENTICKS;
bool			CPS2VM::m_nInVBlank					= false;
CSlot<int>		CPS2VM::m_OnMachineStateChange;
CSlot<int>		CPS2VM::m_OnRunningStateChange;
CSlot<int>		CPS2VM::m_OnNewFrame;
CLogControl		CPS2VM::m_Logging;
CPS2OS*			CPS2VM::m_pOS						= NULL;

//////////////////////////////////////////////////
//Various Message Functions
//////////////////////////////////////////////////

void CPS2VM::CreateGSHandler(GSHANDLERFACTORY pF, void* pParam)
{
	CREATEGSHANDLERPARAM Param;
	if(m_pGS != NULL) return;

	Param.pFactory	= pF;
	Param.pParam	= pParam;

	SendMessage(PS2VM_MSG_CREATEGS, &Param);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::DestroyGSHandler()
{
	if(m_pGS == NULL) return;
	SendMessage(PS2VM_MSG_DESTROYGS);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::CreatePadHandler(PADHANDLERFACTORY pF, void* pParam)
{
	CREATEPADHANDLERPARAM Param;
	if(m_pPad != NULL) return;

	Param.pFactory	= pF;
	Param.pParam	= pParam;

	SendMessage(PS2VM_MSG_CREATEPAD, &Param);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::DestroyPadHandler()
{
	if(m_pPad == NULL) return;
	SendMessage(PS2VM_MSG_DESTROYPAD);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::Resume()
{
	if(m_nStatus == PS2VM_STATUS_RUNNING) return;
	SendMessage(PS2VM_MSG_RESUME);
}

void CPS2VM::Pause()
{
	if(m_nStatus == PS2VM_STATUS_PAUSED) return;
	SendMessage(PS2VM_MSG_PAUSE);
	//This is to make sure everything has been finished and that the thread is really in a sleeping state
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::Initialize()
{
	CreateVM();
	pthread_create(&m_nThread, NULL, EmuThread, NULL);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::Destroy()
{
	SendMessage(PS2VM_MSG_DESTROY);
	DestroyVM();
}

void CPS2VM::SaveState(const char* sPath)
{
	SendMessage(PS2VM_MSG_SAVESTATE, (void*)sPath);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::LoadState(const char* sPath)
{
	SendMessage(PS2VM_MSG_LOADSTATE, (void*)sPath);
	SendMessage(PS2VM_MSG_NULL);
}

void CPS2VM::SendMessage(PS2VM_MSG nMsg, void* pParam)
{
	m_MsgBox.SendMessage(nMsg, pParam);
}

//////////////////////////////////////////////////
//Non extern callable methods
//////////////////////////////////////////////////

void CPS2VM::CreateVM()
{
	printf("PS2VM: Virtual Machine Memory Usage: RAM: %i MBs, BIOS: %i MBs.\r\n", RAMSIZE / 0x100000, BIOSSIZE / 0x100000);
	
	m_pRAM		= (uint8*)malloc(RAMSIZE);
	m_pBIOS		= (uint8*)malloc(BIOSSIZE);

	memset(m_pRAM, 0, RAMSIZE);
	//LoadBIOS();

	m_EE.m_pMemoryMap->InsertReadMap(0x00000000, 0x01FFFFFF, m_pRAM,				MEMORYMAP_TYPE_MEMORY,		0x00);
	m_EE.m_pMemoryMap->InsertReadMap(0x10000000, 0x10FFFFFF, IOPortReadHandler,		MEMORYMAP_TYPE_FUNCTION,	0x01);
	m_EE.m_pMemoryMap->InsertReadMap(0x12000000, 0x12FFFFFF, IOPortReadHandler,		MEMORYMAP_TYPE_FUNCTION,	0x02);
	m_EE.m_pMemoryMap->InsertReadMap(0x1FC00000, 0x1FFFFFFF, m_pBIOS,				MEMORYMAP_TYPE_MEMORY,		0x03);

	m_EE.m_pMemoryMap->InsertWriteMap(0x00000000, 0x01FFFFFF, m_pRAM,				MEMORYMAP_TYPE_MEMORY,		0x00);
	m_EE.m_pMemoryMap->InsertWriteMap(0x10000000, 0x10FFFFFF, IOPortWriteHandler,	MEMORYMAP_TYPE_FUNCTION,	0x01);
	m_EE.m_pMemoryMap->InsertWriteMap(0x12000000, 0x12FFFFFF, IOPortWriteHandler,	MEMORYMAP_TYPE_FUNCTION,	0x02);

	m_EE.m_pArch			= &g_MAEE;
	m_EE.m_pCOP[0]			= &g_COPSCU;
	m_EE.m_pCOP[1]			= &g_COPFPU;

	m_EE.m_pAddrTranslator	= &CMIPS::TranslateAddress64;

	m_EE.m_pTickFunction	= EETickFunction;

	m_nStatus = PS2VM_STATUS_PAUSED;
	
	//Reset subunits
	CSIF::Reset();
	CGIF::Reset();
	CDMAC::Reset();
	CINTC::Reset();

	m_pOS = new CPS2OS;

	printf("PS2VM: Created PS2 Virtual Machine.\r\n");
}

void CPS2VM::DestroyVM()
{
	FREEPTR(m_pRAM);
	FREEPTR(m_pBIOS);
	DELETEPTR(m_pOS);
}

void CPS2VM::SaveVMState(const char* sPath)
{
	CStdStream* pS;
	
	if(m_pGS == NULL)
	{
		printf("PS2VM: GS Handler was not instancied. Cannot save state.\r\n");
		return;
	}

	pS = new CStdStream(fopen(sPath, "wb"));

	pS->Write(&CPS2VM::m_EE.m_State, sizeof(MIPSSTATE));
	pS->Write(CPS2VM::m_pRAM, RAMSIZE);
	m_pGS->SaveState(pS);
	CINTC::SaveState(pS);
	CDMAC::SaveState(pS);

	delete pS;

	printf("PS2VM: Saved state to file '%s'.\r\n", sPath);
}

void CPS2VM::LoadVMState(const char* sPath)
{
	CStdStream* pS;

	if(m_pGS == NULL)
	{
		printf("PS2VM: GS Handler was not instancied. Cannot load state.\r\n");
		return;
	}

	pS = new CStdStream(fopen(sPath, "rb"));

	pS->Read(&CPS2VM::m_EE.m_State, sizeof(MIPSSTATE));
	pS->Read(CPS2VM::m_pRAM, RAMSIZE);
	m_pGS->LoadState(pS);
	CINTC::LoadState(pS);
	CDMAC::LoadState(pS);

	delete pS;

	printf("PS2VM: Loaded state from file '%s'.\r\n", sPath);

	m_OnMachineStateChange.Notify(0);
}

void CPS2VM::LoadBIOS()
{
	CStdStream* pS;
	pS = new CStdStream(fopen("./vfs/rom0/scph10000.bin", "rb"));
	pS->Read(m_pBIOS, BIOSSIZE);
	delete pS;
}

void CPS2VM::IOPortWriteHandler(uint32 nAddress, uint32 nData)
{
	if(nAddress >= 0x10008000 && nAddress <= 0x1000FFFC)
	{
		CDMAC::SetRegister(nAddress, nData);
	}
	else if(nAddress >= 0x12000000 && nAddress <= 0x1200108C)
	{
		m_pGS->WritePrivRegister(nAddress, nData);
	}
	else
	{
		printf("PS2VM: Wrote to an unhandled IO port (0x%0.8X, 0x%0.8X).\r\n", nAddress, nData);
	}
}

uint32 CPS2VM::IOPortReadHandler(uint32 nAddress)
{
	if(nAddress >= 0x10008000 && nAddress <= 0x1000FFFC)
	{
		return CDMAC::GetRegister(nAddress);
	}
	else if(nAddress >= 0x12000000 && nAddress <= 0x1200108C)
	{
		return m_pGS->ReadPrivRegister(nAddress);
	}
	else
	{
		printf("PS2VM: Read an unhandled IO port (0x%0.8X).\r\n", nAddress);
	}
	return 0;
}

unsigned int CPS2VM::EETickFunction(unsigned int nTicks)
{

#ifdef DEBUGGER_INCLUDED

	if(m_EE.m_nIllOpcode != MIPS_INVALID_PC)
	{
		printf("PS2VM: (EmotionEngine) Illegal opcode encountered at 0x%0.8X.\r\n", m_EE.m_nIllOpcode);
		m_EE.m_nIllOpcode = MIPS_INVALID_PC;
	}

#endif

	m_nVBlankTicks--;
	if(m_nVBlankTicks == 0)
	{
		m_nInVBlank = !m_nInVBlank;
		if(m_nInVBlank)
		{
			m_nVBlankTicks = VBLANKTICKS;
			m_pGS->SetVBlank();
			m_pGS->Flip();
			m_pPad->Update();
			m_OnNewFrame.Notify(NULL);
		}
		else
		{
			m_nVBlankTicks = SCREENTICKS;
			m_pGS->ResetVBlank();
		}
	}

#ifdef DEBUGGER_INCLUDED

	if(m_EE.MustBreak())
	{
		return 1;
	}

#endif

	//TODO: Check if we can remove this if there's no debugger around
	if(m_MsgBox.IsMessagePending())
	{
		return 1;
	}

	return 0;
}

void* CPS2VM::EmuThread(void* pParam)
{
	bool nEnd;
	CThreadMsg::MESSAGE Msg;
	nEnd = false;
	while(1)
	{
		if(m_MsgBox.GetMessage(&Msg))
		{
			switch(Msg.nMsg)
			{
			case PS2VM_MSG_PAUSE:
				m_nStatus = PS2VM_STATUS_PAUSED;
				m_OnMachineStateChange.Notify(0);
				m_OnRunningStateChange.Notify(0);
				printf("PS2VM: Virtual Machine paused.\r\n");
				break;
			case PS2VM_MSG_RESUME:
				m_nStatus = PS2VM_STATUS_RUNNING;
				m_OnRunningStateChange.Notify(0);
				printf("PS2VM: Virtual Machine started.\r\n");
				break;
			case PS2VM_MSG_DESTROY:
				DELETEPTR(m_pGS);
				nEnd = true;
				break;
			case PS2VM_MSG_CREATEGS:
				CREATEGSHANDLERPARAM* pCreateGSParam;
				pCreateGSParam = (CREATEGSHANDLERPARAM*)Msg.pParam;
				m_pGS = pCreateGSParam->pFactory(pCreateGSParam->pParam);
				break;
			case PS2VM_MSG_DESTROYGS:
				DELETEPTR(m_pGS);
				break;
			case PS2VM_MSG_CREATEPAD:
				CREATEPADHANDLERPARAM* pCreatePadParam;
				pCreatePadParam = (CREATEPADHANDLERPARAM*)Msg.pParam;
				m_pPad = pCreatePadParam->pFactory(CSIF::GetPadMan(), pCreatePadParam->pParam);
				break;
			case PS2VM_MSG_DESTROYPAD:
				DELETEPTR(m_pPad);
				break;
			case PS2VM_MSG_SAVESTATE:
				SaveVMState((const char*)Msg.pParam);
				break;
			case PS2VM_MSG_LOADSTATE:
				LoadVMState((const char*)Msg.pParam);
				break;
			}
		}
		if(nEnd) break;
		if(m_nStatus == PS2VM_STATUS_PAUSED)
		{
			Sleep(100);
		}
		if(m_nStatus == PS2VM_STATUS_RUNNING)
		{
			RET_CODE nRet;
			if(m_EE.MustBreak())
			{
				nRet = RET_CODE_BREAKPOINT;
			}
			else
			{
				nRet = m_EE.Execute(50000);
			}
			if(nRet == RET_CODE_BREAKPOINT)
			{
				printf("PS2VM: (EmotionEngine) Breakpoint encountered at 0x%0.8X.\r\n", m_EE.m_State.nPC);
				m_nStatus = PS2VM_STATUS_PAUSED;
				m_OnMachineStateChange.Notify(0);
				m_OnRunningStateChange.Notify(0);
				continue;
			}
		}
	}
	return NULL;
}
