#include <stdio.h>
#include <string.h>
#include "RegViewGeneral.h"
#include "PS2VM.h"

#define CLSNAME		_X("CRegViewGeneral")
#define XMARGIN		6

using namespace Framework;

CRegViewGeneral::CRegViewGeneral(HWND hParent, RECT* pR, CMIPS* pC)
{
	if(!DoesWindowClassExist(CLSNAME))
	{
		WNDCLASSEX w;
		memset(&w, 0, sizeof(WNDCLASSEX));
		w.cbSize		= sizeof(WNDCLASSEX);
		w.lpfnWndProc	= CWindow::WndProc;
		w.lpszClassName	= CLSNAME;
		w.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
		w.hInstance		= GetModuleHandle(NULL);
		w.hCursor		= LoadCursor(NULL, IDC_ARROW);
		RegisterClassEx(&w);
	}

	Create(WS_EX_CLIENTEDGE, CLSNAME, _X(""), WS_VISIBLE | WS_CHILD, pR, hParent, NULL);
	SetClassPtr();

	m_pCtx = pC;
	
	m_pOnMachineStateChangeHandler = new CSlotHandlerMethod<CRegViewGeneral, int>(this, &CRegViewGeneral::OnMachineStateChange);
	m_pOnRunningStateChangeHandler = new CSlotHandlerMethod<CRegViewGeneral, int>(this, &CRegViewGeneral::OnRunningStateChange);

	CPS2VM::m_OnMachineStateChange.InsertHandler(m_pOnMachineStateChangeHandler);
	CPS2VM::m_OnRunningStateChange.InsertHandler(m_pOnRunningStateChangeHandler);
}

CRegViewGeneral::~CRegViewGeneral()
{
	CPS2VM::m_OnMachineStateChange.RemoveHandler(m_pOnMachineStateChangeHandler);
	CPS2VM::m_OnRunningStateChange.RemoveHandler(m_pOnRunningStateChangeHandler);
}

void CRegViewGeneral::OnRunningStateChange(int nNothing)
{
	Redraw();
}

void CRegViewGeneral::OnMachineStateChange(int nNothing)
{
	Redraw();
}

HFONT CRegViewGeneral::GetFont()
{
	return CreateFont(-11, 0, 0, 0, 400, 0, 0, 0, 0, 1, 2, 1, 49, _X("Courier New"));
}

void CRegViewGeneral::Paint(HDC hDC)
{
	RECT rwin;
	HFONT nFont;
	xchar sText[256];
	xchar sReg[256];
	MIPSSTATE* s;
	int i;
	unsigned int nX, nY;

	GetClientRect(&rwin);

	BitBlt(hDC, 0, 0, rwin.right, rwin.bottom, NULL, 0, 0, WHITENESS);

	nFont = GetFont();

	SelectObject(hDC, nFont);

	s = &m_pCtx->m_State;
	nY = 3;
	for(i = 0; i < 32; i++)
	{
		xconvert(sReg, CMIPS::m_sGPRName[i], 256);
		xsprintf(sText, _X("%s : 0x%0.8X%0.8X%0.8X%0.8X"), sReg, s->nGPR3[i], s->nGPR2[i], s->nGPR1[i], s->nGPR0[i]);
/*		
		if(i & 1)
		{
			nX = 303;
		}
		else
		{
			nX = 3;
		}
*/

		nX = XMARGIN;

		TextOut(hDC, nX, nY, sText, (int)xstrlen(sText));

//		if(i & 1)
//		{
			nY += 18;
//		}
	}

	xsprintf(sText, _X("LO : 0x%0.8X%0.8X"), s->nLO[1], s->nLO[0]);
	TextOut(hDC, nX, nY, sText, (int)xstrlen(sText));
	nY += 18;

	xsprintf(sText, _X("HI : 0x%0.8X%0.8X"), s->nHI[1], s->nHI[0]);
	TextOut(hDC, nX, nY, sText, (int)xstrlen(sText));
	nY += 18;

	xsprintf(sText, _X("LO1: 0x%0.8X%0.8X"), s->nLO1[1], s->nLO1[0]);
	TextOut(hDC, nX, nY, sText, (int)xstrlen(sText));
	nY += 18;

	xsprintf(sText, _X("HI1: 0x%0.8X%0.8X"), s->nHI1[1], s->nHI1[0]);
	TextOut(hDC, nX, nY, sText, (int)xstrlen(sText));
	nY += 18;

	DeleteObject(nFont);
}
